#include "lib3ds\mesh.h"
#include "lib3ds\material.h"
#include "lib3ds\file.h"
#include "myUtil.h"
#include <math.h>
#include <string.h>
#include <float.h>

// global max/mins
float maxX, maxY, maxZ, minX, minY, minZ;

int GetBodyPart(const char *name)
{
   if (!strcmp(name, "Head"))
      return HEAD;
   else if (!strcmp(name, "Torso"))
      return TORSO;
   else if (!strcmp(name, "Waist"))
      return WAIST;
   else if (!strcmp(name, "RightShoul"))
      return R_SHOULDER;
   else if (!strcmp(name, "RightArm"))
      return R_ARM;
   else if (!strcmp(name, "RightForeA"))
      return R_FOREARM;
   else if (!strcmp(name, "RightHand"))
      return R_HAND;
   else if (!strcmp(name, "RightThumb"))
      return R_THUMB;
   else if (!strcmp(name, "RightIndex"))
      return R_INDEX;
   else if (!strcmp(name, "RightMiddl"))
      return R_MIDDLE;
   else if (!strcmp(name, "RightRing"))
      return R_RING;
   else if (!strcmp(name, "RightPinky"))
      return R_PINKY;
   else if (!strcmp(name, "LeftShould"))
      return L_SHOULDER;
   else if (!strcmp(name, "LeftArm"))
      return L_ARM;
   else if (!strcmp(name, "LeftForeAr"))
      return L_FOREARM;
   else if (!strcmp(name, "LeftHand"))
      return L_HAND;
   else if (!strcmp(name, "LeftThumb"))
      return L_THUMB;
   else if (!strcmp(name, "LeftIndex"))
      return L_INDEX;
   else if (!strcmp(name, "LeftMiddle"))
      return L_MIDDLE;
   else if (!strcmp(name, "LeftRing"))
      return L_RING;
   else if (!strcmp(name, "LeftPinky"))
      return L_PINKY;
   else if (!strcmp(name, "RightThigh"))
      return R_THIGH;
   else if (!strcmp(name, "RightKnee"))
      return R_KNEE;
   else if (!strcmp(name, "RightCalf"))
      return R_CALF;
   else if (!strcmp(name, "RightFoot"))
      return R_FOOT;
   else if (!strcmp(name, "LeftThigh"))
      return L_THIGH;
   else if (!strcmp(name, "LeftKnee"))
      return L_KNEE;
   else if (!strcmp(name, "LeftCalf"))
      return L_CALF;
   else if (!strcmp(name, "LeftFoot"))
      return L_FOOT;
   else if (!strcmp(name, "Coat"))
      return COAT;
   else if (!strcmp(name, "CoatTop"))
      return COAT_TOP;
   else if (!strcmp(name, "LeftSleeve"))
      return L_SLEEVE;
   else if (!strcmp(name, "RightSleev"))
      return R_SLEEVE;
   else
      return -1;
}

void SetOffset(int part)
{
   maxes[part][0] = maxX;
   maxes[part][1] = maxY;
   maxes[part][2] = maxZ;
   mins[part][0] = minX;
   mins[part][1] = minY;
   mins[part][2] = minZ;

   switch(part)
   {
      case COAT:
         printf("Coat\nMax X:%f\nMin X:%f\nMax Y:%f\nMin Y:%f\nMax Z:%f\nMin Z:%f\n",
                maxX, minX, maxY, minY, maxZ, minZ);
         break;
      case COAT_TOP:
         printf("Coat Top\nMax X:%f\nMin X:%f\nMax Y:%f\nMin Y:%f\nMax Z:%f\nMin Z:%f\n",
                maxX, minX, maxY, minY, maxZ, minZ);
         break;
      case L_SLEEVE:
         printf("Left Sleeve\nMax X:%f\nMin X:%f\nMax Y:%f\nMin Y:%f\nMax Z:%f\nMin Z:%f\n",
                maxX, minX, maxY, minY, maxZ, minZ);
         break;
      case R_SLEEVE:
         printf("Right Sleeve\nMax X:%f\nMin X:%f\nMax Y:%f\nMin Y:%f\nMax Z:%f\nMin Z:%f\n",
                maxX, minX, maxY, minY, maxZ, minZ);
         break;
      case L_THIGH:
      case L_CALF:
      case R_THIGH:
      case R_CALF:
      case L_ARM:
      case L_FOREARM:
      case R_ARM:
      case R_FOREARM:
         offsets[part][0] = (maxX + minX) / 2.0f;
         offsets[part][1] = maxY;
         offsets[part][2] = (maxZ + minZ) / 2.0f;
         break;
      case L_FOOT:
      case R_FOOT:
         offsets[part][0] = (maxX + minX) / 2.0f;
         offsets[part][1] = (maxY + minY) / 2.0f;
         offsets[part][2] = (maxZ + minZ) / 6.0f;
         break;
      default:
         offsets[part][0] = (maxX + minX) / 2.0f;
         offsets[part][1] = (maxY + minY) / 2.0f;
         offsets[part][2] = (maxZ + minZ) / 2.0f;
         break;
   }
}

void SetBounds(int part)
{
   float xLen = maxX - minX;
   float yLen = maxY - minY;
   float zLen = maxZ - minZ;

   switch(part)
   {
      case COAT:
      case COAT_TOP:
      case L_SLEEVE:
      case R_SLEEVE:
         break;
      case L_KNEE:
      case R_KNEE:
         bounds[part].center.x = (maxX + minX) / 2.0f;
         bounds[part].center.y = (maxY + minY) / 2.0f;
         bounds[part].center.z = (maxZ + minZ) / 2.0f;

         if (xLen > yLen)
         {
            if (xLen > zLen)
               bounds[part].radius = xLen / 2.0f;
            else
               bounds[part].radius = zLen / 2.0f;
         }
         else if (yLen > zLen)
            bounds[part].radius = yLen / 2.0f;
         else
            bounds[part].radius = zLen / 2.0f;

         bounds[part].height = -1.0;
         bounds[part].radius *= 1.3
            ;
         break;
      case L_THIGH:
      case R_THIGH:
         bounds[part].center.x = (maxX + minX) / 2.0f;
         bounds[part].center.y = (maxY + minY) / 2.0f;
         bounds[part].center.z = (maxZ + minZ) / 2.0f;

         if (xLen > zLen)
            bounds[part].radius = xLen / 2.0f;
         else
            bounds[part].radius = zLen / 2.0f;

         bounds[part].height = yLen;
         bounds[part].radius *= 1.3;
         break;
      default:
         bounds[part].center.x = (maxX + minX) / 2.0f;
         bounds[part].center.y = (maxY + minY) / 2.0f;
         bounds[part].center.z = (maxZ + minZ) / 2.0f;

         if (xLen > zLen)
            bounds[part].radius = xLen / 2.0f;
         else
            bounds[part].radius = zLen / 2.0f;

         bounds[part].height = yLen;
         bounds[part].radius *= 1.2;
         break;
   }
}

void Create3DMesh(const char *fName)
{
   float n1, n2, n3;
   double deno;
   int part = -1;
   
   Lib3dsFile *mFile = lib3ds_file_load(fName);

   if (!mFile)
      return;

   mFile->materials[0].ambient[0];
   
   for (Lib3dsMesh *mesh = mFile->meshes; mesh; mesh = mesh->next)
   {
      if (!strcmp(mesh->name, "Coat") ||
          !strcmp(mesh->name, "CoatTop") ||
          !strcmp(mesh->name, "LeftSleeve") ||
          !strcmp(mesh->name, "RightSleev"))
      {
         part = GetBodyPart(mesh->name);
         std::vector<ClothVertex *> *verts;
         std::vector<ClothFace *> *faces;

         if (part == COAT)
         {
            verts = &coatVerts;
            faces = &coatFaces;
         }
         else if (part == COAT_TOP)
         {
            verts = &coatTopVerts;
            faces = &coatTopFaces;
         }
         else if (part == L_SLEEVE)
         {
            verts = &leftSlVerts;
            faces = &leftSlFaces;
         }
         else if (part == R_SLEEVE)
         {
            verts = &rightSlVerts;
            faces = &rightSlFaces;
         }

         for (int i = 0; i < mesh->points; i++)
            verts->push_back(new ClothVertex());

         for (unsigned int ndx = 0; ndx < mesh->faces; ndx++)
         {
            faces->push_back(new ClothFace());
            Lib3dsFace f = mesh->faceL[ndx];

            for (int pt = 0; pt < 3; pt++)
            {
               ClothVertex *ver = (*verts)[f.points[pt]];

               // set the normals for lighting
               // note that the normals have to be normalized
               n1 = mesh->faceL[ndx].normal[0];
               n2 = mesh->faceL[ndx].normal[1];
               n3 = mesh->faceL[ndx].normal[2];
               deno = sqrt(n1*n1 + n2*n2 + n3*n3);
               ver->ver.normal = Vec3D(n1 / deno, n2 / deno, n3 / deno);
               Lib3dsPoint p = mesh->pointL[f.points[pt]];
               ver->ver.x = ver->home.x = p.pos[0];
               ver->ver.y = ver->home.y = p.pos[1];
               ver->ver.z = ver->home.z = p.pos[2];

               if (ver->ver.x < -0.045162335)
                  ver->front = false;
            }

            ClothVertex *v1 = (*verts)[f.points[0]];
            ClothVertex *v2 = (*verts)[f.points[1]];
            ClothVertex *v3 = (*verts)[f.points[2]];

            float d1 = sqrt((v1->ver.x - v2->ver.x) * (v1->ver.x - v2->ver.x) +
                            (v1->ver.y - v2->ver.y) * (v1->ver.y - v2->ver.y) +
                            (v1->ver.z - v2->ver.z) * (v1->ver.z - v2->ver.z));
            float d2 = sqrt((v1->ver.x - v3->ver.x) * (v1->ver.x - v3->ver.x) +
                            (v1->ver.y - v3->ver.y) * (v1->ver.y - v3->ver.y) +
                            (v1->ver.z - v3->ver.z) * (v1->ver.z - v3->ver.z));
            float d3 = sqrt((v2->ver.x - v3->ver.x) * (v2->ver.x - v3->ver.x) +
                            (v2->ver.y - v3->ver.y) * (v2->ver.y - v3->ver.y) +
                            (v2->ver.z - v3->ver.z) * (v2->ver.z - v3->ver.z));

            if (d1 < 0.5)
            {
               v1->adjacent.push_back(v2);
               v2->adjacent.push_back(v1);
            }
            if (d2 < 0.5)
            {
               v1->adjacent.push_back(v3);
               v3->adjacent.push_back(v1);
            }
            if (d3 < 0.5)
            {
               v2->adjacent.push_back(v3);
               v3->adjacent.push_back(v2);
            }

            (*faces)[ndx]->ndx1 = f.points[0];
            (*faces)[ndx]->ndx2 = f.points[1];
            (*faces)[ndx]->ndx3 = f.points[2];
         }
      }
      else
      {
         if (part > 0)
         {
            SetOffset(part);
            SetBounds(part);
         }

         maxX = maxY = maxZ = -FLT_MAX;
         minX = minY = minZ = FLT_MAX;

         part = GetBodyPart(mesh->name);

         glNewList(part, GL_COMPILE);
         glBegin(GL_TRIANGLES);

         for (unsigned int ndx = 0; ndx < mesh->faces; ndx++)
         {
            Lib3dsFace f = mesh->faceL[ndx];

            for (int pt = 0; pt < 3; pt++)
            {
               // set the normals for lighting
               // note that the normals have to be normalized
               n1 = mesh->faceL[ndx].normal[0];
               n2 = mesh->faceL[ndx].normal[1];
               n3 = mesh->faceL[ndx].normal[2];
               deno = sqrt(n1*n1 + n2*n2 + n3*n3);
               glNormal3f(n1 / deno, n2 / deno, n3 / deno);
               Lib3dsPoint p = mesh->pointL[f.points[pt]];
               glVertex3f(p.pos[0], p.pos[1], p.pos[2]);

               if (p.pos[0] > maxX)
                  maxX = p.pos[0];
               if (p.pos[0] < minX)
                  minX = p.pos[0];
               if (p.pos[1] > maxY)
                  maxY = p.pos[1];
               if (p.pos[1] < minY)
                  minY = p.pos[1];
               if (p.pos[2] > maxZ)
                  maxZ = p.pos[2];
               if (p.pos[2] < minZ)
                  minZ = p.pos[2];
            }
         }

         glEnd();
         glEndList();
      }
   }

   SetOffset(part);
   SetBounds(part);
}