//basic program to read in a mesh file (of .m format from H. Hoppe)
//Hh code modified by ZJW for csc 471

#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <map>
#include <ctime>
#include "myUtil.h"

using namespace std;

#define WORLD_LEN 60

// material definitions
materialStruct BlackShiny = {
  {0.0, 0.0, 0.0, 1.0},
  {0.2, 0.2, 0.2, 1.0},
  {0.2, 0.2, 0.2, 1.0},
  {8.0}
};
materialStruct Default = {
  {0.2, 0.2, 0.2, 1.0},
  {0.8, 0.8, 0.8, 1.0},
  {0.0, 0.0, 0.0, 1.0},
  {8.0}
};

materialStruct RedFlat = {
  {0.3, 0.0, 0.0, 1.0},
  {0.9, 0.0, 0.0, 1.0},
  {0.0, 0.0, 0.0, 1.0},
  {0.0}
};

//camera vectors
Vec3D eye, lookAt, up_vec;

//globals for lighting - use a white light and apply materials
//light position
GLfloat light_pos[4] = {10.0, 3.0, 10.0, 1.0};
GLfloat alt_light_pos[4] = {-10.0, 3.0, -10.0, 1.0};
//light color (ambiant, diffuse and specular)
GLfloat light_amb[4] = {0.6, 0.6, 0.6, 1.0};
GLfloat light_diff[4] = {0.6, 0.6, 0.6, 1.0};
GLfloat light_spec[4] = {0.8, 0.8, 0.8, 1.0};

std::vector<ClothVertex *> coatVerts, coatTopVerts, leftSlVerts, rightSlVerts;
std::vector<ClothFace *> coatFaces, coatTopFaces, leftSlFaces, rightSlFaces;
std::map<int, ClothVertex *> pullList;
std::map<int, ClothFloat> pullDists;

Cylinder bounds[NUM_PARTS];
float offsets[NUM_PARTS][3];
float rots[NUM_PARTS];
float dirs[NUM_PARTS];
float maxes[NUM_PARTS][3];
float mins[NUM_PARTS][3];

//other globals
int GW, GH;
int oldX, oldY;
GLUquadric *quad;
float rot_x, rot_y;
bool showCoat, points;

enum {
   QUIT
};

//initialization calls for opengl for static light
void init_lighting() {
  //turn on light0
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);
  //set up the diffuse, ambient and specular components for the light
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diff);
  glLightfv(GL_LIGHT0, GL_AMBIENT, light_amb);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_spec);
  glLightfv(GL_LIGHT1, GL_DIFFUSE, light_diff);
  glLightfv(GL_LIGHT1, GL_AMBIENT, light_amb);
  glLightfv(GL_LIGHT1, GL_SPECULAR, light_spec);
  //specify our lighting model as 1 normal per face
  glShadeModel(GL_SMOOTH);
}

void pos_light() {
  //set the light's position
  glMatrixMode(GL_MODELVIEW);
  glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
  glLightfv(GL_LIGHT1, GL_POSITION, alt_light_pos);
}

void drawGround()
{
   float len = (float)WORLD_LEN / 2.0;
   float inc = 2.0;

   for (float x = -len; x < len; x += inc)
   {
      for (float z = -len; z < len; z += inc)
      {
         glBegin(GL_QUADS);
            glNormal3f(0.0, 1.0, 0.0);
            glVertex3f(x, 0.0, z);
            glNormal3f(0.0, 1.0, 0.0);
            glVertex3f(x + inc, 0.0, z);
            glNormal3f(0.0, 1.0, 0.0);
            glVertex3f(x + inc, 0.0, z + inc);
            glNormal3f(0.0, 1.0, 0.0);
            glVertex3f(x, 0.0, z + inc);
         glEnd();
      }
   }
}

void reshape(int w, int h) {
   GW = w;
   GH = h;

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(45.0, (float)GW/GH, 0.5, WORLD_LEN * 1.75);
   glMatrixMode(GL_MODELVIEW);
   glViewport(0, 0, w, h);

   glutPostRedisplay();
}

void testArmBounds(int part)
{
   float xOff, yVal, yDist;
   std::vector<ClothVertex *> *verts;
   float top = maxes[part][1];
   float sine = sin(rots[part] * M_PI / 180.0);
   float cosi = cos(rots[part] * M_PI / 180.0);

   if (part == L_ARM)
      verts = &leftSlVerts;
   else
      verts = &rightSlVerts;

   int size = verts->size();

   for (int i = 0; i < size; i++)
   {
      ClothVertex *v = (*verts)[i];

      yDist = top - v->home.y;
      xOff = yDist * sine;
      yVal = yDist * cosi;

      v->ver.x = v->home.x + xOff;
      v->ver.y = top - yVal;
   }
}

void testThighBounds(int part)
{
   float xDist, yDist, zDist, verDist, fX, fY, fZ, temp;
   float maxZ = maxes[part][2];
   float minZ = mins[part][2];
   float maxY = maxes[part][1];
   float minY = mins[part][1] * cos(rots[part] * M_PI / 180.0);

   float tang = tan(rots[part] * M_PI / 180.0);
   float cosi = cos(rots[part] * M_PI / 180.0);
   float sine = sin(rots[part] * M_PI / 180.0);
   int size = coatVerts.size();

   for (int i = 0; i < size; i++)
   {
      ClothVertex *v = coatVerts[i];

      if (v->ver.y > minY && v->ver.y < maxY &&
          v->ver.z > minZ && v->ver.z < maxZ)
      {
         fX = bounds[part].center.x;
         fY = v->ver.y;
         fZ = bounds[part].center.z;
         fX += (maxY - v->ver.y) * tang;
         
         xDist = (v->ver.x - fX) / 2.0;
         fX += xDist;
         fY -= xDist / tang;

         xDist = v->ver.x - fX;
         yDist = v->ver.y - fY;
         zDist = v->ver.z - fZ;

         verDist = sqrt(xDist * xDist + yDist * yDist + zDist * zDist);

         if (verDist < bounds[part].radius)
         {
            if (v->ver.x < fX)
               verDist = verDist - bounds[part].radius;
            else
               verDist = bounds[part].radius - verDist;

            temp = v->ver.x;
            v->ver.x += verDist * cosi;
            pullDists[i * 3].f += v->ver.x - temp;
            temp = v->ver.y;
            v->ver.y += verDist * sine;
            pullDists[i * 3 + 1].f += v->ver.y - temp;
            pullDists[i * 3 + 2].f = 0.0;
            pullList[i] = v;
            v->shifted = true;
         }
      }
   }
}

void testKneeBounds(int part)
{
   float maxY, minY, cosi, sine;;
   float xDist, yDist, zDist, verDist, fX, fY, fZ, temp;
   float maxZ = maxes[part][2];
   float minZ = mins[part][2];

   if (part == L_KNEE)
   {
      cosi = cos(rots[L_THIGH] * M_PI / 180.0);
      sine = -sin(rots[L_THIGH] * M_PI / 180.0);
   }
   else
   {
      cosi = cos(rots[R_THIGH] * M_PI / 180.0);
      sine = -sin(rots[R_THIGH] * M_PI / 180.0);
   }

   maxY = maxes[part][1] * cosi;
   minY = mins[part][1] * cosi;

   int size = coatVerts.size();

   for (int i = 0; i < size; i++)
   {
      ClothVertex *v = coatVerts[i];

      if (v->ver.y > minY && v->ver.y < maxY &&
          v->ver.z > minZ && v->ver.z < maxZ)
      {
         fX = bounds[part].center.x + bounds[part].center.y * sine;
         fY = bounds[part].center.y * cosi;
         fZ = bounds[part].center.z;

         xDist = v->ver.x - fX;
         yDist = v->ver.y - fY;
         zDist = v->ver.z - fZ;

         verDist = sqrt(xDist * xDist + yDist * yDist + zDist * zDist);

         if (verDist < bounds[part].radius)
         {
            if (v->ver.x < fX)
               verDist = verDist - bounds[part].radius;
            else
               verDist = bounds[part].radius - verDist;

            xDist = verDist * cosi;
            // prevent the knees from dragging the back of the coat forward
            if (!v->front && xDist > 0.0)
               return;

            temp = v->ver.x;
            v->ver.x += xDist;
            pullDists[i * 3].f += v->ver.x - temp;
            temp = v->ver.y;
            v->ver.y += verDist * sine;
            pullDists[i * 3 + 1].f += v->ver.y - temp;
            pullDists[i * 3 + 2].f = 0.0;
            pullList[i] = v;
            v->shifted = true;
         }
      }
   }
}

void testCalfBounds(int part)
{
   float maxY, minY, tcosi, tsine, cosi, sine;
   float xDist, yDist, zDist, verDist, fX, fY, fZ, temp;
   float maxZ = maxes[part][2];
   float minZ = mins[part][2];
   int thigh;

   if (part == L_CALF)
      thigh = L_THIGH;
   else
      thigh = R_THIGH;

   tcosi = cos(rots[thigh] * M_PI / 180.0);
   tsine = -sin(rots[thigh] * M_PI / 180.0);

   cosi = cos(rots[part] * M_PI / 180.0);

   if (rots[thigh] > 0.0)
      sine = -sin(rots[part] * M_PI / 180.0);
   else
      sine = sin(rots[part] * M_PI / 180.0);

   maxY = maxes[part][1] * tcosi;
   minY = mins[part][1] * tcosi * cosi;

   float tang = tan(rots[part] * M_PI / 180.0);
   int size = coatVerts.size();

   for (int i = 0; i < size; i++)
   {
      ClothVertex *v = coatVerts[i];

      if (v->ver.y > minY && v->ver.y < maxY &&
          v->ver.z > minZ && v->ver.z < maxZ)
      {
         fX = bounds[part].center.x + maxY * tsine + (maxY - bounds[part].center.y) * sine;
         fY = v->ver.y;
         fZ = bounds[part].center.z;
         fX += (maxY - v->ver.y) * tang;
         
         xDist = (v->ver.x - fX) / 2.0;
         fX += xDist;
         fY -= xDist / tang;

         xDist = v->ver.x - fX;
         yDist = v->ver.y - fY;
         zDist = v->ver.z - fZ;

         verDist = sqrt(xDist * xDist + yDist * yDist + zDist * zDist);

         if (verDist < bounds[part].radius)
         {
            if (v->ver.x < fX)
               verDist = verDist - bounds[part].radius;
            else
               verDist = bounds[part].radius - verDist;

            temp = v->ver.x;
            v->ver.x += verDist * cosi;
            pullDists[i * 3].f += v->ver.x - temp;
            temp = v->ver.y;
            v->ver.y += verDist * sine;
            pullDists[i * 3 + 1].f += v->ver.y - temp;
            pullDists[i * 3 + 2].f = 0.0;
            pullList[i] = v;
            v->shifted = true;
         }
      }
   }
}

void walk(int thigh, int calf, int foot, int arm, int forearm)
{
   // thigh
   if (rots[thigh] >= MAX_THIGH_ROT)
      dirs[thigh] = -1.0;
   else if (rots[thigh] <= MIN_THIGH_ROT)
      dirs[thigh] = 0.5;
   else if (rots[thigh] < 0.0 && dirs[thigh] < 0.0)
      dirs[thigh] = -0.5;
   else if (rots[thigh] > 0.0 && dirs[thigh] > 0.0)
      dirs[thigh] = 1.0;

   // calf
   if (dirs[thigh] > 0.0)
   {
      if (rots[thigh] > 0.75 * MAX_THIGH_ROT)
         dirs[calf] = 1.0;
      else if (rots[thigh] > 0.0)
         dirs[calf] = -1.0;
      else// if (rots[thigh] < 0.0)
         dirs[calf] = 1.0;
   }
   else
   {
      if (rots[thigh] > 0.5 * MAX_THIGH_ROT)
         dirs[calf] = 1.0;
      else if (rots[thigh] > 0.0)
         dirs[calf] = 0.0;
      else// if (rots[thigh] < 0.0)
         dirs[calf] = -1.0;
   }

   if (equalf(rots[thigh], 0.0))
      rots[thigh] = rots[calf] = 0.0;

   rots[thigh] += dirs[thigh];
   rots[calf] += dirs[calf];

   if (rots[thigh] < 0.25 * MAX_THIGH_ROT && rots[thigh] > 0.25 * MIN_THIGH_ROT)
      rots[foot] = -(rots[thigh] + rots[calf]);
   else
      rots[foot] = 0.0;

   // arm
   dirs[arm] = dirs[thigh] / -4.0;
   rots[arm] += dirs[arm];

   if (rots[arm] < -5.0)
      rots[arm] = 5.0;

   // forearm
   if (rots[thigh] < 0.0)
   {
      if (dirs[thigh] < 0.0)
         dirs[forearm] = 0.5;
      else
         dirs[forearm] = -0.5;

      rots[forearm] += dirs[forearm];
   }
   else
   {
      dirs[forearm] = 0.0;
      rots[forearm] = 0.0;
   }

   testThighBounds(thigh);

   if (thigh == R_THIGH)
      testKneeBounds(R_KNEE);
   else if (thigh == L_THIGH)
      testKneeBounds(L_KNEE);

   testCalfBounds(calf);
   //testBounds(forearm);

   std::vector<ClothVertex *> cleanList;
   std::vector<ClothVertex *>::iterator iter, end;
   std::map<int, ClothVertex *>::iterator vIter = pullList.begin();
   std::map<int, ClothVertex *>::iterator vEnd = pullList.end();
   std::map<int, ClothFloat>::iterator dIter = pullDists.begin();
   float fX, fY, fZ;

   for ( ; vIter != vEnd; vIter++)
   {
      fX = dIter->second.f;
      dIter++;
      fY = dIter->second.f;
      dIter++;
      fZ = dIter->second.f;
      dIter++;

      vIter->second->Pull(0.001, fX, fY, fZ, &cleanList);

      for (iter = cleanList.begin(), end = cleanList.end(); iter != end; iter++)
         (*iter)->canPull = true;

      cleanList.clear();
   }

   for (vIter = pullList.begin(); vIter != vEnd; vIter++)
      vIter->second->shifted = false;

   pullList.clear();
   pullDists.clear();

   testArmBounds(arm);
}

void animate(int data)
{
   walk(L_THIGH, L_CALF, L_FOOT, L_ARM, L_FOREARM);
   walk(R_THIGH, R_CALF, R_FOOT, R_ARM, R_FOREARM);

   glutPostRedisplay();
   glutTimerFunc(10, animate, 0);
}

void drawCoat()
{
   int size = coatFaces.size();

   glBegin(GL_TRIANGLES);
   for (int i = 0; i < size; i++)
   {
      ClothFace *face = coatFaces[i];
      ClothVertex *v1 = coatVerts[face->ndx1];
      ClothVertex *v2 = coatVerts[face->ndx2];
      ClothVertex *v3 = coatVerts[face->ndx3];

      glNormal3f(v1->ver.normal.x, v1->ver.normal.y, v1->ver.normal.z);
      glVertex3f(v1->ver.x, v1->ver.y, v1->ver.z);
      glNormal3f(v2->ver.normal.x, v2->ver.normal.y, v2->ver.normal.z);
      glVertex3f(v2->ver.x, v2->ver.y, v2->ver.z);
      glNormal3f(v3->ver.normal.x, v3->ver.normal.y, v3->ver.normal.z);
      glVertex3f(v3->ver.x, v3->ver.y, v3->ver.z);
   }
   glEnd();

   size = coatTopFaces.size();

   glBegin(GL_TRIANGLES);
   for (int i = 0; i < size; i++)
   {
      ClothFace *face = coatTopFaces[i];
      ClothVertex *v1 = coatTopVerts[face->ndx1];
      ClothVertex *v2 = coatTopVerts[face->ndx2];
      ClothVertex *v3 = coatTopVerts[face->ndx3];

      glNormal3f(v1->ver.normal.x, v1->ver.normal.y, v1->ver.normal.z);
      glVertex3f(v1->ver.x, v1->ver.y, v1->ver.z);
      glNormal3f(v2->ver.normal.x, v2->ver.normal.y, v2->ver.normal.z);
      glVertex3f(v2->ver.x, v2->ver.y, v2->ver.z);
      glNormal3f(v3->ver.normal.x, v3->ver.normal.y, v3->ver.normal.z);
      glVertex3f(v3->ver.x, v3->ver.y, v3->ver.z);
   }
   glEnd();

   size = leftSlFaces.size();

   glBegin(GL_TRIANGLES);
   for (int i = 0; i < size; i++)
   {
      ClothFace *face = leftSlFaces[i];
      ClothVertex *v1 = leftSlVerts[face->ndx1];
      ClothVertex *v2 = leftSlVerts[face->ndx2];
      ClothVertex *v3 = leftSlVerts[face->ndx3];

      glNormal3f(v1->ver.normal.x, v1->ver.normal.y, v1->ver.normal.z);
      glVertex3f(v1->ver.x, v1->ver.y, v1->ver.z);
      glNormal3f(v2->ver.normal.x, v2->ver.normal.y, v2->ver.normal.z);
      glVertex3f(v2->ver.x, v2->ver.y, v2->ver.z);
      glNormal3f(v3->ver.normal.x, v3->ver.normal.y, v3->ver.normal.z);
      glVertex3f(v3->ver.x, v3->ver.y, v3->ver.z);
   }
   glEnd();

   size = rightSlFaces.size();

   glBegin(GL_TRIANGLES);
   for (int i = 0; i < size; i++)
   {
      ClothFace *face = rightSlFaces[i];
      ClothVertex *v1 = rightSlVerts[face->ndx1];
      ClothVertex *v2 = rightSlVerts[face->ndx2];
      ClothVertex *v3 = rightSlVerts[face->ndx3];

      glNormal3f(v1->ver.normal.x, v1->ver.normal.y, v1->ver.normal.z);
      glVertex3f(v1->ver.x, v1->ver.y, v1->ver.z);
      glNormal3f(v2->ver.normal.x, v2->ver.normal.y, v2->ver.normal.z);
      glVertex3f(v2->ver.x, v2->ver.y, v2->ver.z);
      glNormal3f(v3->ver.normal.x, v3->ver.normal.y, v3->ver.normal.z);
      glVertex3f(v3->ver.x, v3->ver.y, v3->ver.z);
   }
   glEnd();
}

void display()
{  
   materials(RedFlat);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glMatrixMode(GL_MODELVIEW);
    
   glPushMatrix();
      //set up the camera
      gluLookAt(eye.x, eye.y, eye.z, //eye position
                lookAt.x, lookAt.y, lookAt.z, //point we are looking at
                up_vec.x, up_vec.y, up_vec.z);  //up vector

      //position the lights in the scene
      pos_light();

      //drawGround();

      if (showCoat)
      {
         if (points)
            glPolygonMode(GL_FRONT_AND_BACK, GL_POINT);

         drawCoat();
      }

      if (points)
         glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

      glCallList(HEAD);
      glCallList(TORSO);
      glCallList(WAIST);
      glCallList(R_SHOULDER);
      glCallList(L_SHOULDER);

      glPushMatrix();
         glTranslatef(offsets[R_ARM][0], offsets[R_ARM][1], offsets[R_ARM][2]);
         glRotatef(rots[R_ARM], 0.0, 0.0, 1.0);
         glTranslatef(-offsets[R_ARM][0], -offsets[R_ARM][1], -offsets[R_ARM][2]);
         glCallList(R_ARM);

         glPushMatrix();
            glTranslatef(offsets[R_FOREARM][0], offsets[R_FOREARM][1], offsets[R_FOREARM][2]);
            glRotatef(rots[R_FOREARM], 0.0, 0.0, 1.0);
            glTranslatef(-offsets[R_FOREARM][0], -offsets[R_FOREARM][1], -offsets[R_FOREARM][2]);
            glCallList(R_FOREARM);

            glPushMatrix();
               glCallList(R_HAND);
               glCallList(R_THUMB);
               glCallList(R_INDEX);
               glCallList(R_MIDDLE);
               glCallList(R_RING);
               glCallList(R_PINKY);
            glPopMatrix();
         glPopMatrix();
      glPopMatrix();

      glPushMatrix();
         glTranslatef(offsets[L_ARM][0], offsets[L_ARM][1], offsets[L_ARM][2]);
         glRotatef(rots[L_ARM], 0.0, 0.0, 1.0);
         glTranslatef(-offsets[L_ARM][0], -offsets[L_ARM][1], -offsets[L_ARM][2]);
         glCallList(L_ARM);

         glPushMatrix();
            glTranslatef(offsets[L_FOREARM][0], offsets[L_FOREARM][1], offsets[L_FOREARM][2]);
            glRotatef(rots[L_FOREARM], 0.0, 0.0, 1.0);
            glTranslatef(-offsets[L_FOREARM][0], -offsets[L_FOREARM][1], -offsets[L_FOREARM][2]);
            glCallList(L_FOREARM);

            glPushMatrix();
               glCallList(L_HAND);
               glCallList(L_THUMB);
               glCallList(L_INDEX);
               glCallList(L_MIDDLE);
               glCallList(L_RING);
               glCallList(L_PINKY);
            glPopMatrix();
         glPopMatrix();
      glPopMatrix();
 
      glPushMatrix();
         glTranslatef(offsets[R_THIGH][0], offsets[R_THIGH][1], offsets[R_THIGH][2]);
         glRotatef(rots[R_THIGH], 0.0, 0.0, 1.0);
         glTranslatef(-offsets[R_THIGH][0], -offsets[R_THIGH][1], -offsets[R_THIGH][2]);
         glCallList(R_THIGH);
         glCallList(R_KNEE);

         glPushMatrix();
            glTranslatef(offsets[R_CALF][0], offsets[R_CALF][1], offsets[R_CALF][2]);
            glRotatef(rots[R_CALF], 0.0, 0.0, 1.0);
            glTranslatef(-offsets[R_CALF][0], -offsets[R_CALF][1], -offsets[R_CALF][2]);
            glCallList(R_CALF);

            glPushMatrix();
               glTranslatef(offsets[R_FOOT][0], offsets[R_FOOT][1], offsets[R_FOOT][2]);
               glRotatef(rots[R_FOOT], 0.0, 0.0, 1.0);
               glTranslatef(-offsets[R_FOOT][0], -offsets[R_FOOT][1], -offsets[R_FOOT][2]);
               glCallList(R_FOOT);
            glPopMatrix();
         glPopMatrix();
      glPopMatrix();

      glPushMatrix();
         glTranslatef(offsets[L_THIGH][0], offsets[L_THIGH][1], offsets[L_THIGH][2]);
         glRotatef(rots[L_THIGH], 0.0, 0.0, 1.0);
         glTranslatef(-offsets[L_THIGH][0], -offsets[L_THIGH][1], -offsets[L_THIGH][2]);
         glCallList(L_THIGH);
         glCallList(L_KNEE);

         glPushMatrix();
            glTranslatef(offsets[L_CALF][0], offsets[L_CALF][1], offsets[L_CALF][2]);
            glRotatef(rots[L_CALF], 0.0, 0.0, 1.0);
            glTranslatef(-offsets[L_CALF][0], -offsets[L_CALF][1], -offsets[L_CALF][2]);
            glCallList(L_CALF);

            glPushMatrix();
               glTranslatef(offsets[L_FOOT][0], offsets[L_FOOT][1], offsets[L_FOOT][2]);
               glRotatef(rots[L_FOOT], 0.0, 0.0, 1.0);
               glTranslatef(-offsets[L_FOOT][0], -offsets[L_FOOT][1], -offsets[L_FOOT][2]);
               glCallList(L_FOOT);
            glPopMatrix();
         glPopMatrix();
      glPopMatrix();

   glPopMatrix();
   glFlush();
   
   int size = coatVerts.size();

   for (int i = 0; i < size; i++)
      coatVerts[i]->Animate();

   glutSwapBuffers(); 
}

//the mouse button callback
void mouse(int button, int state, int x, int y)
{
   if (button == GLUT_LEFT_BUTTON)
   {
      if (state == GLUT_DOWN) /* if the left button is clicked */
      {
         // invert the y
         y = GH - 1 - y;
         oldX = x;
         oldY = y;

         glutPostRedisplay();
      }
   }
}

//the mouse move callback
void mouseMove(int x, int y) {
   float deltaX, deltaY, len;
   Vec3D tempE, tempL, gaze;

   // invert the y
   y = GH - 1 - y;

   // calculate the rotation amount
   deltaX = x - oldX;
   deltaY = y - oldY;
   rot_x += (float)(deltaX / GW) * M_PI;
   rot_y += (float)(deltaY / GH) * M_PI;

   // bind y rotation to +/- 05
   float bind = 50.0 * M_PI / 180.0;
   if (rot_y > bind)
      rot_y = bind;
   else if (rot_y < -bind)
      rot_y = -bind;

   // calculate the translation from the origin
   tempE = eye;
   tempL = lookAt;
   tempE.x -= eye.x;
   tempE.y -= eye.y;
   tempE.z -= eye.z;
   tempL.x -= eye.x;
   tempL.y -= eye.y;
   tempL.z -= eye.z;
   gaze = tempL - tempE;
   len = gaze.getLength();
   tempE.x = len * cos(rot_y) * cos(rot_x);
   tempE.y = len * sin(rot_y);
   tempE.z = len * cos(rot_y) * cos((M_PI / 2.0) - rot_x);

   // apply translation to lookAt
   lookAt.x += tempE.x - tempL.x;
   lookAt.y += tempE.y - tempL.y;
   lookAt.z += tempE.z - tempL.z;

   oldX = x;
   oldY = y;
   glutPostRedisplay();
}

//the keyboard callback
void keyboard(unsigned char key, int x, int y )
{
   Vec3D w_vec, u_vec, v_vec, gaze = lookAt - eye;
   float inc, len = gaze.getLength();
   w_vec.x = -gaze.x / len;
   w_vec.y = -gaze.y / len;
   w_vec.z = -gaze.z / len;
   u_vec = w_vec.crossProduct(up_vec);
   len = u_vec.getLength();
   u_vec.x /= len;
   u_vec.y /= len;
   u_vec.z /= len;

   switch (key)
   {
      case 'a': case 'A' :
         inc = (float)WORLD_LEN / 40.0;
         eye.x += inc * u_vec.x;
         eye.y += inc * u_vec.y;
         eye.z += inc * u_vec.z;
         lookAt.x += inc * u_vec.x;
         lookAt.y += inc * u_vec.y;
         lookAt.z += inc * u_vec.z;
         break;
      case 'd': case 'D' :
         inc = (float)WORLD_LEN / 40.0;
         eye.x -= inc * u_vec.x;
         eye.y -= inc * u_vec.y;
         eye.z -= inc * u_vec.z;
         lookAt.x -= inc * u_vec.x;
         lookAt.y -= inc * u_vec.y;
         lookAt.z -= inc * u_vec.z;
         break;
      case 's': case 'S' :
         inc = (float)WORLD_LEN / 40.0;
         eye.x += inc * w_vec.x;
         eye.y += inc * w_vec.y;
         eye.z += inc * w_vec.z;
         lookAt.x += inc * w_vec.x;
         lookAt.y += inc * w_vec.y;
         lookAt.z += inc * w_vec.z;
         break;
      case 'w': case 'W' :
         inc = (float)WORLD_LEN / 40.0;
         eye.x -= inc * w_vec.x;
         eye.y -= inc * w_vec.y;
         eye.z -= inc * w_vec.z;
         lookAt.x -= inc * w_vec.x;
         lookAt.y -= inc * w_vec.y;
         lookAt.z -= inc * w_vec.z;
         break;
      case 'c': case 'C' :
         showCoat = !showCoat;
         break;
      case 'q': case 'Q' :
         exit( EXIT_SUCCESS );
         break;
      case 'h': case 'H' :
         printf("3D World Program:\n");
         printf("Click and drag the mouse to look around.\n");
         printf("Press 'w' to move forward.\n");
         printf("Press 's' to move backward.\n");
         printf("Press 'a' to move to the left.\n");
         printf("Press 'd' to move to the right.\n");
         printf("Press 'q' to quit the program.\n");

         break;
      case 'p': case 'P' :
         points = !points;
   }

   glutPostRedisplay();
}

int main( int argc, char** argv )
{
   //set up my window
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
   glutInitWindowSize(300, 300); 
   glutInitWindowPosition(100, 100);
   glutCreateWindow("Mesh display");
   glClearColor(0.0, 0.0, 0.0, 1.0);
  
   //register glut callback functions
   glutDisplayFunc( display );
   glutReshapeFunc( reshape );
   glutMouseFunc( mouse );
   glutMotionFunc( mouseMove );
   glutKeyboardFunc( keyboard );
   glutTimerFunc(10, animate, 0);

   //enable z-buffer
   glEnable(GL_DEPTH_TEST);
   glEnable(GL_LIGHTING);
   glPointSize(3.0);
   //glEnable(GL_CULL_FACE);
   //glEnable(GL_TEXTURE_2D);
   //glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
   //glCullFace(GL_BACK);

   //quad = gluNewQuadric();
   //gluQuadricDrawStyle(quad, GLU_FILL);
   //gluQuadricNormals(quad, GLU_SMOOTH);
   //gluQuadricTexture(quad, GL_TRUE);
   //initialize_tombstones();
   //initialize_vertices();

   //glEnable(GL_NORMALIZE);
   //load in the other image textures
   //LoadTexture("deadgrass256.bmp", 1);
   //LoadTexture("tombstone256.bmp", 2);
   //LoadTexture("jellyfish256.bmp", 3)

   for (int i = 0; i < NUM_PARTS; i++)
   {
      rots[i] = 0.0;
      dirs[i] = 1.0;
   }

   dirs[R_THIGH] = -0.5;

   //camera
   rot_x = rot_y = 0.0;
   eye.x = eye.y = eye.z = 0.0;
   float len = 10.0;
   lookAt.x = len * cos(rot_y) * cos(rot_x);
   lookAt.y = len * sin(rot_y);
   lookAt.z = len * cos(rot_y) * cos((M_PI / 2.0) - rot_x);
   eye.x = -50.0;
   eye.y = 0.0;
   eye.z = (float)WORLD_LEN / -4.0;
   lookAt.x += eye.x;
   lookAt.y += eye.y;
   lookAt.z += eye.z;

   up_vec.x = 0.0;
   up_vec.y = 1.0;
   up_vec.z = 0.0;

   init_lighting();
   Create3DMesh("gambit.3ds");
   showCoat = true;
   points = false;
   glutMainLoop();
}