#include "myUtil.h"

// Calculate the normal for a face and contribute to the face's vertices
void CalculateNormal(Vertex *vx1, Vertex *vx2, Vertex *vx3, Vertex *vx4)
{
   Vec3D vec1, vec2, normal;

   vec1.origin[0] = vec2.origin[0] = vx1->x;
   vec1.origin[1] = vec2.origin[1] = vx1->y;
   vec1.origin[2] = vec2.origin[2] = vx1->z;
   vec1.x = vx2->x - vec1.origin[0];
   vec1.y = vx2->y - vec1.origin[1];
   vec1.z = vx2->z - vec1.origin[2];
   vec2.x = vx3->x - vec2.origin[0];
   vec2.y = vx3->y - vec2.origin[1];
   vec2.z = vx3->z - vec2.origin[2];
   
   normal = vec1.crossProduct(vec2);

   // Normalize the normal
   float len = normal.getLength();
   normal.x /= len;
   normal.y /= len;
   normal.z /= len;

   // Contribute to vertices' normals
   vx1->normal += normal;
   vx2->normal += normal;
   vx3->normal += normal;

   if (vx4)
      vx4->normal += normal;
}

// transform functions
void apply_translation(float &xTrans, float &yTrans, float fX, float fY)
{
   float left = 2.0 * GW/GH;
   float right = -2.0 * GW/GH;
   float top = 2.0;
   float bottom = -2.0;

   xTrans += fX;
   yTrans += fY;
}

void apply_rotation(GLfloat *rotMatrix, float theta, float xAxis, float yAxis,
                    float zAxis)
{
   glPushMatrix();
      glLoadIdentity();
      glRotatef(theta, xAxis, yAxis, zAxis);
      glMultMatrixf(rotMatrix);
      glGetFloatv(GL_MODELVIEW_MATRIX, rotMatrix);
   glPopMatrix();
}

void apply_scale(float &scale, float fX, float fY, float fZ)
{
   scale += fY;
}

void materials(materialStruct materials) {
  glMaterialfv(GL_FRONT, GL_AMBIENT, materials.ambient);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, materials.diffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR, materials.specular);
  glMaterialfv(GL_FRONT, GL_SHININESS, materials.shininess);
}