/**
 * myUtil.h
 *
 * A set of utility functions and globals for CSC 471
 *
 * Erik Kitson
 * Spring 06
 **/
#ifndef MY_UTIL_H
#define MY_UTIL_H

#include <assert.h>
#include <fstream>
#include <GL\glut.h>
#include <math.h>
#include <vector>

#define M_PI       3.14159265358979323846
#define NUM_PARTS  34
#define MAX_THIGH_ROT 40.0
#define MIN_THIGH_ROT -20.0

void Create3DMesh(const char *fName);
extern float offsets[NUM_PARTS][3];
extern float maxes[NUM_PARTS][3];
extern float mins[NUM_PARTS][3];
//global variables - due to GLUT
//global width and height
extern int GW;
extern int GH;
//current mouse position in pixel coordinate
extern int x;
extern int y;

enum {
   HEAD = 1,
   TORSO,
   WAIST,
   R_SHOULDER,
   R_ARM,
   R_FOREARM,
   R_HAND,
   R_THUMB,
   R_INDEX,
   R_MIDDLE,
   R_RING,
   R_PINKY,
   L_SHOULDER,
   L_ARM,
   L_FOREARM,
   L_HAND,
   L_THUMB,
   L_INDEX,
   L_MIDDLE,
   L_RING,
   L_PINKY,
   R_THIGH,
   R_KNEE,
   R_CALF,
   R_FOOT,
   L_THIGH,
   L_KNEE,
   L_CALF,
   L_FOOT,
   COAT,
   COAT_TOP,
   L_SLEEVE,
   R_SLEEVE
};

// helper functions
static bool equalf(float f1, float f2)
{
   return fabs(f1 - f2) < .001;
}

// for converting coordinates
static float p2i_x(int x)
{
	float g = (float)GW / 2;
	float h = ((float)GW - 1) / 2;
   return ((float)x - h) / g;
}

static float p2i_y(int y)
{
   float i = (float)GH / 2;
   float j = ((float)GH - 1) / 2;
   return ((float)y - j) / i;
}

static float i2w_x(float x)
{
   float r = (float)GW / (float)GH;
   float l = -(float)GW / (float)GH;
   float c = 2 / (r - l);
   float d = (-r - l) / (r - l);
   return (x - d) / c;
}

static float i2w_y(float y)
{
   float t = 1;
   float b = -1;
   float e = 2 / (t - b);
   float f = (-t - b) / (t - b);
   return (y - f) / e;
}

// a structure to store 2D points
typedef struct vector2 {
   float x;
   float y;
} vector2;

// a structure to store 3D vectors
typedef struct Vec3D {
   float x;
   float y;
   float z;
   float origin[3];
   float color[3];

   Vec3D() :
      x(0.0),
      y(0.0),
      z(0.0)
   {
      origin[0] = origin[1] = origin[2] = 0.0;
      color[0] = 0.0;
      color[1] = 1.0;
      color[1] = 0.5;
   }

   Vec3D(float fX, float fY, float fZ) :
      x(fX),
      y(fY),
      z(fZ)
   {
      origin[0] = origin[1] = origin[2] = 0.0;
      color[0] = 0.0;
      color[1] = 1.0;
      color[1] = 0.5;
   }

   Vec3D &operator=(Vec3D &other)
   {
      if (&other != this)
      {
         x = other.x;
         y = other.y;
         z = other.z;
         color[0] = other.color[0];
         color[1] = other.color[1];
         color[2] = other.color[2];
         origin[0] = other.origin[0];
         origin[1] = other.origin[1];
         origin[2] = other.origin[2];
      }

      return *this;
   }

   Vec3D &operator+=(Vec3D &other)
   {
      x += other.x;
      y += other.y;
      z += other.z;

      return *this;
   }

   Vec3D operator-(const Vec3D &vec)
   {
      return Vec3D(x - vec.x, y - vec.y, z - vec.z);
   }

   float dotProduct(Vec3D &other)
   {
      return x * other.x + y * other.y + z * other.z;
   }

   Vec3D crossProduct(Vec3D &other)
   {
      Vec3D cross;

      cross.origin[0] = origin[0];
      cross.origin[1] = origin[1];
      cross.origin[2] = origin[2];
      cross.x = y * other.z - z * other.y;
      cross.y = z * other.x - x * other.z;
      cross.z = x * other.y - y * other.x;

      float dot1 = dotProduct(cross);
      float dot2 = other.dotProduct(cross);
      //printf("Dot Product 1:%d\n", dot1);
      //printf("Dot Product 2:%d\n", dot2);
      assert(equalf(dot1, 0.0));
      assert(equalf(dot2, 0.0));
      return cross;
   }

   void projectZ()
   {
      float lengthTest = sqrt(pow(x, 2) + pow(y, 2));

      if (lengthTest > 1.0)
      {
         // normalize
         x = x / sqrt(pow(x, 2) + pow(y, 2));
         y = y / sqrt(pow(x, 2) + pow(y, 2));
         z = 0;
      }
      else
      {
         // inside the unit sphere
         z = sqrt(1.0 - (pow(x, 2) + pow(y, 2)));
      }
   }

   float getLength()
   {
      return sqrt(pow(x, 2) + pow(y, 2) + pow(z, 2));
   }

   Vec3D getOrthogonal2D()
   {
      Vec3D orth(x, y, z);
      orth.origin[0] = origin[0] + (x / 2.0);
      orth.origin[1] = origin[1] + (y / 2.0);
      orth.color[0] = 0.5;
      orth.color[1] = 0.5;

      float length = getLength();
      orth.x /= length;
      orth.y /= length;
      orth.z /= length;

      float tempX = orth.x;
      orth.x = -orth.y;
      orth.y = tempX;

      return orth;
   }

   void Draw()
   {
      glPointSize(2.0);

      glBegin(GL_POINTS);
         glColor3f(0.0, 0.0, 0.0);
         glVertex3f(origin[0], origin[1], origin[2]);
      glEnd();

      glBegin(GL_LINES);
         glColor3f(color[0], color[1], color[2]);
         glVertex3f(origin[0], origin[1], origin[2]);
         glVertex3f(origin[0] + x, origin[1] + y, origin[2] + z);
      glEnd();
   }
} Vec3D;

// very simple data structure to 3D vertices
typedef struct Vertex {
  float x;
  float y;
  float z;
  Vec3D normal;

  Vertex(float in_x, float in_y, float in_z)
     : x(in_x), y(in_y), z(in_z)
  {
  }

  Vertex()
  {
     normal.x = normal.y = normal.z = 0.0;
     x = y = z = 0.0;
  }
} Vertex;

typedef struct ClothVertex {
   Vertex ver;
   Vertex home;
   std::vector<ClothVertex *> adjacent;
   bool canPull;
   bool shifted;
   bool front;

   ClothVertex()
   {
      canPull = true;
      shifted = false;
      front = true;
   }

   void Animate()
   {
      if (!equalf(ver.x, home.x))
         ver.x -= (ver.x - home.x) / 20.0;
      if (!equalf(ver.y, home.y))
         ver.y -= (ver.y - home.y) / 20.0;
      if (!equalf(ver.z, home.z))
         ver.z -= (ver.z - home.z) / 20.0;
   }

   void Pull(float factor, float fX, float fY, float fZ,
             std::vector<ClothVertex *> *list)
   {
      if (canPull && fX + fY + fZ > 0.0001)
      {
         canPull = false;
         fX /= 2.0;
         fY /= 2.0;
         fZ /= 2.0;

         if (!shifted)
         {
            ver.x += fX * factor;
            ver.y += fY * factor;
            ver.z += fZ * factor;
         }

         factor *= 3.0;
         list->push_back(this);
         int verts = adjacent.size();

         for (int i = 0; i < verts; i++)
         {
            adjacent[i]->Pull(factor, fX, fY, fZ, list);
         }
      }
   }
} ClothVertex;

typedef struct ClothFace {
   float ndx1;
   float ndx2;
   float ndx3;
} ClothFace;

typedef struct ClothFloat {
   float f;

   ClothFloat()
   {
      f = 0.0;
   }
} ClothFloat;

// simple structure to store material info
typedef struct materialStruct {
  GLfloat ambient[4];
  GLfloat diffuse[4];
  GLfloat specular[4];
  GLfloat shininess[1];
} materialStruct;

// bounding struct
typedef struct Cylinder {
   Vertex center;
   float height;     // a height of -1 indicates a sphere
   float radius;

   Cylinder()
   {
      height = radius = 0.0;
   }
} Cylinder;

extern Cylinder bounds[NUM_PARTS];
extern std::vector<ClothVertex *> coatVerts, coatTopVerts, leftSlVerts, rightSlVerts;
extern std::vector<ClothFace *> coatFaces, coatTopFaces, leftSlFaces, rightSlFaces;

// Calculate the normal for a face and contribute to the face's vertices
void CalculateNormal(Vertex *vx1, Vertex *vx2, Vertex *vx3, Vertex *vx4);
// transform functions
void apply_translation(float &xTrans, float &yTrans, float fX, float fY);
void apply_rotation(GLfloat *rotMatrix, float theta, float xAxis, float yAxis,
                    float zAxis);
void apply_scale(float &scale, float fX, float fY, float fZ);
//sets up a specific material
void materials(materialStruct materials);

#endif